from dataclasses import Field, dataclass, fields
from inspect import getmodule
from textwrap import dedent
from typing import Any, Dict, List, Type, get_args

from pytest_embrace.anno import Comment

from .embrace import registry
from .exc import EmbraceError


class EmbraceTestGenError(EmbraceError):
    ...


@dataclass
class AnnotationInfo:
    type: Any
    annotations: List[Any]


AnnotationMap = Dict[str, AnnotationInfo]


def _get_pep_593_values(cls: Type) -> AnnotationMap:
    out: AnnotationMap = {}
    for k, v in cls.__annotations__.items():
        attr_name = getattr(v, "__name__", getattr(v.__class__, "__name__", ""))
        if attr_name not in {"Annotated", "_AnnotatedAlias"}:
            continue
        typ, *annotations = get_args(v)
        out[k] = AnnotationInfo(typ, annotations)
    return out


def _stringify_type(type: Type) -> str:
    return str(type) if not repr((type)).startswith("<class") else type.__name__


def _field_to_cute_type_hint(field: Field, anno_map: AnnotationMap) -> str:
    type_hint = _stringify_type(field.type)
    if field.name not in anno_map:
        return f"{field.name}: {type_hint}"

    anno_info = anno_map[field.name]
    comment = next((v for v in anno_info.annotations if isinstance(v, Comment)), None)
    type_hint = _stringify_type(anno_info.type)
    text = f"{field.name}: {type_hint}"
    if comment is not None:
        text += f"  # {comment.text}"
    return text


def gen_text(name: str, table: bool = False) -> str:
    case_type = registry().get(name)
    if case_type is None:
        raise EmbraceTestGenError(f"No such test type '{name}'.")

    anno_map = _get_pep_593_values(case_type)

    type_hints = "\n".join(
        _field_to_cute_type_hint(f, anno_map) for f in fields(case_type)
    )

    source = getmodule(case_type)
    case_import = (
        dedent(
            f"""
            import {case_type.__name__} from {mod_name}
            """
        )
        if (mod_name := getattr(source, "__name__", None)) is not None
        else ""
    )

    # not using dedent bc newlines in the type hints are hard
    return f"""
from pytest_embrace import CaseArtifact
{case_import}

{type_hints}


def test({name}: CaseArtifact[{case_type.__name__}]) -> None:
    ...
            """



from dataclasses import dataclass
from typing import Any, Generic, Protocol, TypeVar

CaseType = TypeVar("CaseType")
CoCaseType = TypeVar("CoCaseType", contravariant=True)


@dataclass
class CaseArtifact(Generic[CoCaseType]):
    case: CoCaseType
    actual_result: Any


CaseArtifactType = TypeVar("CaseArtifactType", bound=CaseArtifact, covariant=True)


class CaseRunner(Protocol[CoCaseType, CaseArtifactType]):
    def __call__(self, case: CoCaseType) -> CaseArtifactType:
        ...


class Trickle:
    """This type is used to identify attributes marked with trickles()."""


def trickles() -> Any:
    """Marks an attribute as one that can 'trickle down' from module scope
    into table cases as a default value."""
    return Trickle()



from __future__ import annotations

from collections.abc import Iterator, MutableMapping
from dataclasses import dataclass
from functools import partial
from inspect import signature
from typing import Callable, Dict, Generic, Type, TypeVar

import pytest

from .case import CaseRunner, CaseType
from .exc import CaseConfigurationError

RegistryValue = Type["CaseType"]

T = TypeVar("T")


@dataclass
class CaseTypeInfo(Generic[T]):
    cls: T


class OneTimeOnlyMapping(MutableMapping):
    def __init__(self) -> None:
        self._mapping: Dict[str, RegistryValue] = {}

    def __getitem__(self, key: str) -> RegistryValue:
        return self._mapping[key]

    def __setitem__(self, key: str, val: RegistryValue) -> None:
        if key not in self._mapping:
            self._mapping[key] = val
            return

        existing = self._mapping[key]

        if type(existing) == type(val):
            return

        raise CaseConfigurationError(
            f"Already registered a case fixture called for {key} for {val}."
        )

    def __iter__(self) -> Iterator[str]:
        return iter(self._mapping)

    def __len__(self) -> int:
        return len(self._mapping)

    def __delitem__(self, _: str) -> None:
        raise NotImplementedError


_registry = OneTimeOnlyMapping()


class Embrace(Generic[CaseType]):
    # _runner_registry: ClassVar[OneTimeOnlyMapping] = OneTimeOnlyMapping()

    def __init__(self, case_type: Type[CaseType]):
        self.case_type = case_type
        self.wrapped_func: CaseRunner | None = None
        self.runner: partial | None = None
        # _registry[name] = self.case_type

    def register_case_runner(
        self, func: CaseRunner
    ) -> Callable[[pytest.FixtureRequest], partial]:

        self.wrapped_func = func

        @pytest.fixture
        def run_case_fixture(request: pytest.FixtureRequest) -> partial:
            sig = signature(func)
            kwargs = {
                name: request.getfixturevalue(name)
                for name, param in sig.parameters.items()
                if param.annotation != self.case_type
            }
            self.runner = partial(func, **kwargs)

            return self.runner

        return run_case_fixture

    def caller_fixture_factory(
        self, name: str
    ) -> Callable[[pytest.FixtureRequest, CaseType], None]:
        _registry[name] = self.case_type
        self.caller_fixture_name = name

        @pytest.fixture
        def caller_fixture(request: pytest.FixtureRequest, case: CaseType) -> None:
            assert self.wrapped_func is not None
            # calling `getfixturevalue` for the wrapped func gets `run_case_fixture`.
            # doing *that* assigns self.runner at the last possible moment
            request.getfixturevalue(self.wrapped_func.__name__)  # type: ignore
            assert self.runner is not None
            self.runner(case)

        return caller_fixture


def registry() -> OneTimeOnlyMapping:
    return _registry



from __future__ import annotations

from dataclasses import Field, asdict, dataclass, field, fields, is_dataclass
from types import ModuleType
from typing import Any, Dict, List, Type, Union

from pydantic import create_model
from pydantic.error_wrappers import ValidationError
from pydantic.types import StrictBool, StrictBytes, StrictFloat, StrictInt, StrictStr

# from . import anno
from .case import CaseType, Trickle
from .exc import CaseConfigurationError


class Sentinel:
    ...


@dataclass
class AttrInfo:
    dc_field: Field
    module_value: Any
    name: str = field(init=False)

    def __post_init__(self) -> None:
        self.name = self.dc_field.name


ShouldBecomeStrictBuiltinTypes = Union[str, bytes, int, float, bool]
StrictPydanticTypes = Union[StrictStr, StrictBytes, StrictInt, StrictFloat, StrictBool]

PYDANTIC_STRICTIFICATION_MAP = {
    str: StrictStr,
    bytes: StrictBytes,
    int: StrictInt,
    float: StrictFloat,
    bool: StrictBool,
}


def _strictify(
    t: Any,
) -> Union[Type[StrictPydanticTypes], Type[ShouldBecomeStrictBuiltinTypes], Type[Any]]:
    return PYDANTIC_STRICTIFICATION_MAP.get(t, t)


class ModuleInfo:
    def __init__(self, *, cls: Type[CaseType], mod: ModuleType):
        defined = dir(mod)
        spec = fields(cls)
        validator_kwargs: Dict[str, Any] = {}
        self.attrs: Dict[str, AttrInfo] = {}

        for field_ in spec:
            if field_.name not in defined:
                continue

            if field_.name == "table":
                raise CaseConfigurationError(
                    "You've chosen the only illegal property name!"
                    " It's used by the framwork. Please say 'table_' instead."
                )

            if field_.name == "should":
                # 'should' gets special treatment
                pass

            module_value = getattr(mod, field_.name)
            attr_info = AttrInfo(field_, module_value)
            self.attrs[attr_info.name] = attr_info

            validator_kwargs[field_.name] = (
                _strictify(attr_info.dc_field.type),
                module_value,
            )

        Validator = create_model(f"{mod.__name__}__ModuleValidator", **validator_kwargs)
        Validator(**self.kwargs())

    def kwargs(self) -> Dict[str, Any]:
        return {a.name: a.module_value for a in self.attrs.values()}


def _raise_non_dataclass(o: object) -> None:
    if not is_dataclass(o):
        raise CaseConfigurationError("Must use a dataclass for case object.")


def _report_validation_error(exc: ValidationError, *, target_name: str) -> None:
    errors = exc.errors()
    errors_disambiguation = "\n".join(
        f"    Variable/Arg '{e['loc'][0]}'"
        f" should be of type {e['type'].lstrip('type_error.')}"
        for e in errors
    )
    raise CaseConfigurationError(
        f"{len(errors)} invalid attr values in '{target_name}':\n"
        f"{errors_disambiguation}"
    ) from exc


def from_module(cls: Type[CaseType], module: ModuleType) -> CaseType:
    _raise_non_dataclass(cls)

    try:
        info = ModuleInfo(mod=module, cls=cls)
    except ValidationError as e:
        _report_validation_error(e, target_name=module.__name__)

    try:
        return cls(**info.kwargs())
    except TypeError:
        raise CaseConfigurationError(
            f"Incorrect or missing attributes in module {module.__name__}."
            f"\nGot attributes {set(info.kwargs())}"
        )


class PydanticConfig:
    arbitrary_types_allowed = True


def revalidate_dataclass(case: CaseType, *, alias: str) -> CaseType:
    _raise_non_dataclass(case)
    kwargs = asdict(case)
    validator_kwargs: Dict[str, Any] = {
        k: (_strictify(type(v)), v) for k, v in kwargs.items()
    }

    Validator = create_model(
        f"{case.__class__.__name__}__CaseValidator",
        **validator_kwargs,
        __config__=PydanticConfig,  # type: ignore
    )

    try:
        Validator(**kwargs)
    except ValidationError as e:
        _report_validation_error(e, target_name=alias)

    return case


def handle_table_trickle_down(
    cls: Type[CaseType], module: ModuleType
) -> List[CaseType]:
    table = getattr(module, "table", None)
    assert table is not None

    trickle_attr_defaults = {
        attr: getattr(module, attr)
        for attr in dir(cls)
        if not attr.startswith("__") and isinstance(getattr(cls, attr), Trickle)
    }
    # annotated_module_attrs = {k: pep593[k] for k in dir(module) if k in pep593}

    out: List[CaseType] = []
    for i, case_ in enumerate(table):
        case = revalidate_dataclass(case_, alias=f"{cls.__name__} #{i + 1}")
        for k, v in asdict(case).items():
            if isinstance(v, Trickle):
                setattr(case, k, trickle_attr_defaults[k])

        out.append(case)

    return out



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytest_embrace']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.9.1,<2.0.0', 'pyperclip>=1.8.2,<2.0.0', 'pytest>=7.0,<8.0']

extras_require = \
{':python_version == "3.8"': ['typing-extensions>=4.2.0,<5.0.0']}

entry_points = \
{'console_scripts': ['embrace = pytest_embrace.cli:main'],
 'pytest11': ['pytest_embrace = pytest_embrace.plugin']}

setup_kwargs = {
    'name': 'pytest-embrace',
    'version': '0.3.0',
    'description': 'Use modern Python type hints to design expressive tests. Reject boilerplate. Embrace complexity.',
    'long_description': '# pytest-embrace :gift_heart:\n\nThe `pytest-embrace` plugin enables judicious, repeatable, lucid unit testing.\n\n## Philosophy :bulb:\n\n1. Table-oriented (parametrized) tests are indespensible.\n2. Type hints and modern Python dataclasses are very good.\n3. Language-level APIs (like namespaces) are a honkin\' great idea.\n4. Code generation is *really* underrated.\n5. The wave of type-driven Python tools like Pydantic and Typer (both dependencies of this library) is very cowabunga––and only just beginning :ocean:\n\n## Features :white_check_mark:\n\n- [x] Completely customizable test design\n- [x] Type hints everywhere\n- [x] Table-oriented testing\n- [x] Strongly-typed test namespaces\n- [x] Highly cusomizable code generation––powered by Pep 593\n- [x] Readable errors, early and often\n\n## Basic Usage :wave:\n\nLike any pytest plugin, `pytest-embrace` is configured in `conftest.py`.\n\nThe main ingredients are:\n\n1. The "case" –– which can be any class decorated with `builtins.dataclasses.dataclass`.\n2. The "runner" –– which is just a tricked out Pytest fixture to run assertions against your case.\n3. The "caller" –– which is is another tricked out fixture that your tests will use.\n\n```python\n# conftest.py\nfrom dataclasses import dataclass\nfrom typing import Callable\n\nfrom pytest_embrace import CaseArtifact, Embrace\n\n\n@dataclass\nclass Case:\n    arg: str\n    func: Callable\n    expect: str\n\n\nembrace = Embrace(Case)\n\n@embrace.register_case_runner\ndef run_simple(case: Case):\n    result = case.func(case.arg)\n    assert result == case.expect\n    return result\n\n\nsimple_case = embrace.caller_fixture_factory(\'simple_case\')\n```\n\nWith the above conftest, you can write tests like so:\n\n1. Make a module with attributes matching your `Embrace()`\'d object\n2. Request your caller fixture in normal Pytest fashion\n\n```python\n# test_func.py\narg = \'Ainsley\'\nfunc = lambda x: x * 2\nexpect = \'AinsleyAinsley\'\n\n\ndef test(simple_case):\n\t...\n```\n\nOr you can go table-oriented and run many tests from one module––just like with `pytest.mark.parametrize`.\n\n```python\n# test_many_func.py\nfrom conftest import Case\n\ntable = [\n    Case(arg="haha", func=lambda x: x.upper(), expect="HAHA"),\n    Case(arg="wow damn", func=lambda x: len(x), expect=8),\n    Case(arg="sure", func=lambda x: hasattr(x, "beep"), expect=False),\n]\n\n\ndef test(simple_case):\n    ...\n```\n\n### Strongly Typed Namespaces :muscle:\n\nBefore even completing the setup phase of your `Embrace()`\'d tests, this plugin uses [Pydantic](https://pydantic-docs.helpmanual.io/) to validate the values set in your test modules. No type hints required.\n\nThat means there\'s no waiting around for expensive setups before catching simple mistakes.\n\n```python\n# given this case...\narg = "Ainsley"\nfunc = lambda x: x * 2\nexpect = b"AinsleyAinsley"\n\n\ndef test(simple_case):\n    ...\n```\n\nRunning the above immediately produces this error:\n\n```python\nE   pytest_embrace.exc.CaseConfigurationError: 1 invalid attr values in module \'test_wow\':\nE       Variable \'expect\' should be of type str\n```\n\nThe auxilary benefit of this feature is hardening the design of your code\'s interfaces––even interfaces that exist beyond the "vanishing point" of incoming data that you can\'t be certain of: Command line inputs, incoming HTTP requests, structured file inputs, etc.\n\n## Code Generation :robot:\n\nInstalling `pytest-embrace` adds a flag to `pytest` called `--embrace`.\n\nIt can be used to scaffold tests based on any of your registered cases.\n\nWith the example from above, you can do this out of the box:\n\n```shell\npytest --embrace simple_case\n```\n\nWhich puts this in your clipboard:\n\n```python\n# test_more.py\nfrom pytest_embrace import CaseArtifact\nfrom conftest import Case\n\narg: str\nfunc: "Callable"\nexpect: str\n\n\ndef test(simple_case: CaseArtifact[Case]):\n    ...\n```\n\nCopypasta\'d test cases like this can also be table-style: [Soon.]\n\n```shell\npytest --embrace-table 3\n```\n\nThe value passed to the `--table` flag will produce that many rows.\n\n```python\n# test_table_style.py\nfrom pytest_embrace import CaseArtifact\nfrom conftest import Case\n\ntable = [\n    # Case(arg=..., func=..., expect=...),\n    # Case(arg=..., func=..., expect=...),\n    # Case(arg=..., func=..., expect=...),\n]\n\ndef test(simple_case: CaseArtifact[Case]):\n    ...\n```\n\nBy default, each item is commented out so you don\'t end up with linter errors upon opening your new file.\n\nIf that\'s not cool, don\'t worry! It\'s configurable. :sunglasses:\n\n### Config With Pep 593 :star2:\n\nIn order to customize the behavior of your test cases, `pytest-embrace` :flushed: embraces :flushed:  the new `Annotated` type.\n\n> :information_source:\n> If you\'ve never heard of Pep 593 or `Annotated`, the **tl;dr** is that `Annotated[<type>, ...]` takes any number of arguments after the first one (the actual hint) that developers (me) can use at rumtime.\n\nThe `pytest_embrace.anno` namespace provides a number of utilities for controlling test parsing and code generation via `Annotated`.\n\nHere\'s an example of using `anno.Comment` to put comments in generated test suites:\n\n```python\nfrom dataclasses import dataclass\n\nfrom pytest_embrace import Embrace, anno\n\n\n@dataclass\nclass AnnotatedCase:\n    name: Annotated[str, anno.Comment("This is ... pretty cool.")]\n\n\nembrace = Embrace(AnnotatedCase)\n\n\n@embrace.register_case_runner\ndef run(case: AnnotatedCase, fix: str) -> None:\n    pass\n\n\nanno_case = embrace.caller_fixture_factory("anno_case")\n```\n\nCalling the generation utility...\n\n```shell\npytest --embrace anno_case\n```\n\nGets you this:\n\n```python\nfrom pytest_embrace import CaseArtifact\n\nimport AnnotatedCase from conftest\n\n\nname: str  # This is ... pretty cool.\n\n\ndef test(anno_case: CaseArtifact[AnnotatedCase]) -> None:\n    ...\n```\n',
    'author': 'Ainsley McGrath',
    'author_email': 'mcgrath.ainsley@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import pytest
from pyperclip import copy

from .embrace import registry
from .exc import EmbraceError
from .gen import gen_text
from .loader import from_module, handle_table_trickle_down


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    registry_ = registry()
    embracers = [name for name in metafunc.fixturenames if name in registry_]

    if len(embracers) > 1:
        raise EmbraceError(f"Can't request multiple Embrace fixtures. Got {embracers}")

    if len(embracers) == 0:
        return

    (sut,) = embracers
    cls = registry_[sut]
    if hasattr(metafunc.module, "table"):
        cases = handle_table_trickle_down(cls, metafunc.module)
    else:
        cases = [from_module(cls, metafunc.module)]

    metafunc.parametrize("case", cases, ids=[str(c) for c in cases])


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption("--embrace", help="")
    parser.addoption("--embrace-table")
    parser.addoption("--embrace-doc")


def pytest_runtestloop(session: pytest.Session) -> object:
    generate_for = session.config.getoption("--embrace")
    if generate_for is None:
        return None

    registry_ = registry()
    if generate_for not in registry_:
        pytest.exit(
            f"No such fixture '{generate_for}'."
            f" Your options are {sorted([*registry_])}"
        )

    copypasta = gen_text(generate_for)
    print(f"\nCopying the following output to your clipboard:\n{copypasta}")
    copy(copypasta)

    return object()  # stop the loop
